import { curry } from '../curry.js'
import { Suspend, suspend, pure } from './free.js'

/** @import {InferredMorphism, Morphism, ChainConstructor} from './types.js' */

/**
 * @template T
 * @param {T} x
 * @returns {x}
 */
export const id = x => x

export const compose = curry(
  /**
   * @template T, U, V
   * @param {Morphism<U, V>} f
   * @param {Morphism<T, U>} g
   * @param {T} x
   * @returns V
   */
  (f, g, x) => f(g(x))
)

export const kleisliCompose = curry(
  /**
   * @template {ChainConstructor<T>} M, T, U
   * @param {M} f
   * @param {Morphism<T, U>} g
   * @param {T} x
   * @returns U
   */
  (f, g, x) => f(x).chain(g)
)

/**
 * @template T
 * @param {T} x
 * @returns {Suspend<T>}
 */
export const liftF = x => suspend(
  x,
  /** @type {InferredMorphism<T>} */(pure)
)

