import { id } from './fn.js'
import { Algebra, Monad } from './index.js'

/** @import { InferredMorphism } from './types.js' */

/**
 * @template T
 * @typedef {InferredMorphism<T>} ReaderFn
 */

/** @template T */
export class Reader extends Algebra(Monad) {
  #run

  /**
   * @param {InferredMorphism<T>} run 
   */
  constructor(run) {
    super()

    this.#run = run
  }

  /**
   * @template T
   * @param {T} a
   * @returns {Reader<T>}
   */
  static of(a) {
    return new Reader(/** @type {InferredMorphism<T>} */(_env => a))
  }

  /** @template T */
  static ask() {
    return new Reader(/** @type {InferredMorphism<T>} */(id))
  }

  /**
   * @param {InferredMorphism<T>} f 
   * @returns {Reader<T>}
   */
  map(f) {
    return new Reader(/** @type {InferredMorphism<T>} */(env => f(this.#run(env))))
  }

  /**
   * @param {InferredMorphism<T>} f 
   * @returns {Reader<T>}
   */
  chain(f) {
    return new Reader(env => f(this.#run(env)).run(env))
  }

  /**
   * @param {InferredMorphism<T>} f 
   * @returns {Reader<T>}
   */
  local(f) {
    return new Reader(/** @type {InferredMorphism<T>} */(env => this.#run(f(env))))
  }

  /**
   * @template R
   * @param {T} env 
   * @returns {R}
   */
  run(env) {
    return this.#run(env)
  }
}

